/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:12:00 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include "def.h"
#include "unldef.h"

extern str20 pr;
extern bool verbose;
extern bool X;
extern int feature_type;
extern bool outAscii;
extern int N;
extern str100 featNameFile;
extern str100 featFile;

extern Pixel	**TP;
extern complex	**Harmonics;
extern float	**Amplitudes;

int nrFeat;

Pixel		**P = NULL, **Q = NULL;
int		NPoints;
CPoint		*Points = NULL;



static str80 buf;


bool is_2_power( x )
int x;
{
 if( x == 2 || x == 4 || x == 8 || x == 16 || x == 32 || x == 64 || x == 128 ||
x == 256 || x == 512 || x == 1024 || x == 2048 || x == 4096 ) return(TRUE);
 return(FALSE);
}


Pixel **alloc_pattern( dim )
int dim;
{
 Pixel **P = NULL;
 int i;

 P = (Pixel**) malloc( dim * sizeof(Pixel*) );
 for( i = 0; i < dim; i++ )
 {
   P[i] = (Pixel*) malloc( dim * sizeof(Pixel) );
   CHKPTR(P[i]);
 }
 CHKPTR(P);
 return( P );
}


void free_pattern( P, dim )
Pixel **P;
int dim;
{
 int i;

 for( i = 0; i < dim; i++ )
   FREE(P[i]);
 FREE(P);
}


void Alloc_Orig_Pat()
{
 P = alloc_pattern( N );
 Q = alloc_pattern( N );
 TP = alloc_pattern( N );
 Points = (CPoint*) malloc( N*N*sizeof(struct CPoint_));
 CHKPTR(Points);
}


void Free_Orig_Pat()
{
 free_pattern( P, N );
 free_pattern( Q, N );
 free_pattern( TP, N );
 FREE(Points);
}


void dumpFeat()
{
 switch( feature_type )
 {
   case FOURIER_DESCRIPTORS:	dumpFeat_FD(); break;
   case UFF:			dumpFeat_UFF(); break;
   default: EXIT("Unknown feature type");
 }
}


void calcFeat( C )
Pixel **C;
{
 copyPattern( C, Q );
 clearPattern( TP );
 DisplayPats( Q, TP );
 thin_Zhang_Suen( Q );
 m_connectivity( Q );
 DisplayPats( Q, TP );

 switch( feature_type )
 {
   case FOURIER_DESCRIPTORS: calc_1D_Fourier_Descriptors( Q ); break; /**/
   case UFF:
     UNL_Transform( Q, TP );
     DisplayPats( Q, TP );
     thin_Zhang_Suen_Polar( TP );
     m_connectivity_Polar( TP );
     DisplayPats( Q, TP );


     /* If you calculate ALL UNL Fourier Features, use FFT */
     prepareFFT( TP, Harmonics );
     FFT( Harmonics, Amplitudes );
     NormFFT( Harmonics, Amplitudes );


     /* If you calculate only a few UNL Fourier Features, use FT */
     /* example_calc_UFF();	/**/
     break;
   default: EXIT("Unknown feature type");
 }
}


static void gen_feats()
{
 switch( feature_type )
 {
   case UFF:
     Alloc_UNL_Trans();
     gen_uff_feat_desc();
     get_pat();
     Free_UNL_Trans();
     break;
   case FOURIER_DESCRIPTORS:
     Alloc_FD();
     Alloc_Orig_Pat();
     gen_fd_feat_desc();
     get_pat();
     Free_Orig_Pat();
     Free_FD();
     break;
   default: EXIT("Unknown feature type");
 }
}


int main( argc, argv )
int argc;
char **argv;
{
 parse_input( argc, argv );
 init_all();

 gen_feats();

#ifdef XWINDOW
 if( X )
   { printf("%sHit enter to exit...", pr ); gets(buf); }
#endif
 close_all();
 return(1);
}
