/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:55 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "def.h"
#include "unldef.h"

extern int N, NU;
extern int             NPolarPoints;
extern PolarPoint       *PolarPoints;

#ifdef PI
#undef PI
#endif
#define PI		3.14159265358979323846

#define DIRECT_FFT 	-1
#define INV_FFT    	 1

static float  *SinTab = NULL, *CosTab = NULL;
static int *BitRev = NULL;

/****************************************************/
/* function mirror just transform the binary        */
/* representation of an integer into its mirror     */
/* binary representation on maxlevel bits           */
/* eg : with maxlevel = 6,  000101 becomes 101000   */
/****************************************************/
static int mirror(index,maxlevel)
int index;
int maxlevel;
{
  register int i;
  int res;

  res = 0;
  for (i=0;i<maxlevel;i++)
    {
       res = (res << 1) | (index & (int)1);
       index = (index>> 1);
    }
  return(res);
}

/****************************************************/
/* Auxiliary tables for sins and cosins		    */
/* 'Inv_Direc' -1 FFT     1 Inv. FFT		    */
/****************************************************/
static void InitTables(Inv_Direc)
 int Inv_Direc;
{ int i,j,interlevel; float Omega;

 SinTab = (float*) malloc( N * sizeof(float) ); CHKPTR(SinTab);
 CosTab = (float*) malloc( N * sizeof(float) ); CHKPTR(CosTab);

 interlevel = (int)(log((double) N) / log((double) 2));
 Omega = Inv_Direc * 2 * PI / (float)N;   /* FFT */
 for (i=0; i< N; i++) {
   BitRev[i] = mirror(i,interlevel);
   SinTab[i]= sin(Omega * i);
   CosTab[i]= cos(2*PI*i/(float)N);   /* cos(a) = cos(-a) */
 }
}

/*****************************************************/
/* recursive procedure to calculate the FFT on 2^N   */
/*****************************************************/
static void RecFFT(A,k,l,p,q,Ind_Omg)
complex **A;
int k,l; /* indexes for the begin and the end on the lines taken */
int p,q; /* indexes for the begin and the end of the columns taken */
int Ind_Omg;
{
 int i,j,m;
 complex amb,a_b,cmd,c_d;
 int argument[3];

 if (k != l)
   {
      m = (l-k)/2;

      for (i = k;i < k + m;i++)
	   {
	     argument[0] = Ind_Omg * (i-k);
	     for (j = p;j < p + m;j++)
	       {
	         argument[1] = Ind_Omg * (j-p);
	         argument[2] = Ind_Omg * (i-k + j-p);

		         /* a + b */
              amb.Re = A[i][j].Re + A[i+m][j].Re;
              amb.Im = A[i][j].Im + A[i+m][j].Im;
		         /* a - b */
              a_b.Re = A[i][j].Re - A[i+m][j].Re;
              a_b.Im = A[i][j].Im - A[i+m][j].Im;
		         /* c + d */
	          cmd.Re = A[i][j+m].Re + A[i+m][j+m].Re;
	          cmd.Im = A[i][j+m].Im + A[i+m][j+m].Im;
		         /* c - d */
	          c_d.Re = A[i][j+m].Re - A[i+m][j+m].Re;
	          c_d.Im = A[i][j+m].Im - A[i+m][j+m].Im;

	          A[i][j].Re = amb.Re + cmd.Re;
	          A[i][j].Im = amb.Im + cmd.Im;

	          A[i+m][j].Re = 
	              (a_b.Re + c_d.Re)*CosTab[argument[0]] -
		           (a_b.Im + c_d.Im)*SinTab[argument[0]];
	          A[i+m][j].Im = 
	             (a_b.Re + c_d.Re)*SinTab[argument[0]] +
		           ( a_b.Im + c_d.Im)*CosTab[argument[0]];
	          A[i][j+m].Re = 
	             (amb.Re - cmd.Re)*CosTab[argument[1]] -
		           ( amb.Im - cmd.Im)*SinTab[argument[1]];
	          A[i][j+m].Im = 
	             (amb.Re - cmd.Re)*SinTab[argument[1]] +
		           (amb.Im - cmd.Im)*CosTab[argument[1]];
	          A[i+m][j+m].Re = 
	             (a_b.Re - c_d.Re)*CosTab[argument[2]] -
		           (a_b.Im - c_d.Im)*SinTab[argument[2]];
	          A[i+m][j+m].Im = 
	             (a_b.Re - c_d.Re)*SinTab[argument[2]] +
		           (a_b.Im - c_d.Im)*CosTab[argument[2]];
	        }
	}

    if (m != 0)
  	  {
	    RecFFT(A,k,k+m,p,p+m,Ind_Omg * 2);
	    RecFFT(A,k+m,l,p,p+m,Ind_Omg * 2);
	    RecFFT(A,k,k+m,p+m,q,Ind_Omg * 2);
	    RecFFT(A,k+m,l,p+m,q,Ind_Omg * 2);
	  }
  }
}

/********************************************************/
/* FFT_2D is a procedure which produce the inverse of   */
/* Fourier Transformation.                              */
/********************************************************/
static void FFT_2D(A,interlevel)
complex **A;
int interlevel;
{
  int dimension;
  int i,j,i1,j1;
  float coeff,Omega;

  dimension = (1 << interlevel);

  RecFFT(A,0,dimension,0,dimension,1);

  for (i=0;i<dimension; i++)
    for (j=0;j<dimension;j++)
      { A[i][j].Re /= (float)dimension; A[i][j].Im /= (float)dimension; }

    /* swap every rows i with its mirror image */
  for (i=0;i<=(dimension - 1);i++)
   {
      i1 = BitRev[i];
      if (i1 > i)   /* if i1 <= i, the swapping have been already made */
       for (j=0;j<dimension;j++)
	     {  
	       coeff = A[i][j].Re;
	       A[i][j].Re = A[i1][j].Re;
	       A[i1][j].Re = coeff;

   	      /* this may be remove if you are sure that the result is a  *
	       * matrix of real which means that the original matrix must *
	       * verify the following conditions :                        *
	       *               ______              ______                 *
	       *  A(N-i,N-j) = A(i,j) , A(0,N-j) = A(0,j)                 *
	       *             ______                                       *
	       *  A(N-i,0) = A(i,0) for i,j > 0    and A(0,0) is a pure   *
	       *  real                                                    */

	       coeff = A[i][j].Im;
	       A[i][j].Im = A[i1][j].Im;
	       A[i1][j].Im = coeff;
	     }
    }

  for (j=0;j<=(dimension - 1);j++) /* idem for the columns */
    {
      j1 = BitRev[j];
      if (j1 >j)
       for (i=0;i<dimension;i++)
	     {
	       coeff = A[i][j].Re;
	       A[i][j].Re = A[i][j1].Re;
	       A[i][j1].Re = coeff;

	       /* part which may be removed if (see above)   */
	       coeff = A[i][j].Im;
	       A[i][j].Im = A[i][j1].Im;
	       A[i][j1].Im = coeff;
	     }
    }
}

/***************************************************** InitFFT *** ***********
    Procedure to be compatible with the one called by the application
 ******************************************************************************/
void initFFT()
{ 
 BitRev = (int*) malloc( N * sizeof(int) ); CHKPTR( BitRev );
 InitTables( DIRECT_FFT );   /* Init Sin Cosin tables  */
}


void closeFFT()
{
 FREE( SinTab ); FREE( CosTab );
 FREE( BitRev );
}

/***************************************************** TwoDimFT ***************
    Procedure to be compatible with the one called by the application
 ******************************************************************************/
#define TEST_
void FFT( Harmonics, Amplitudes )
complex **Harmonics;
float **Amplitudes;
{
 int u, v;
 float Re, Im;

 FFT_2D( Harmonics, (int)(log((double) N) / log((double) 2)) );

 for(u=0;u<N;u++)
   for(v=0;v<NU;v++) 
   {
     Re = Harmonics[u][v].Re;
     Im = Harmonics[u][v].Im;
     Amplitudes[u][v] = (float) sqrt((double)(Re*Re+Im*Im));
#ifdef TEST
     {
       float phase, mag, norm, F00;
       if( u == 3 && (v == 0) )
       {
         F00 = Amplitudes[0][0];
         phase = (float) atan2( Im, Re );
         mag = Amplitudes[u][v];
         norm = Amplitudes[u][v] / Amplitudes[0][0];
         printf("Re:%7.3f(%7.3f) Im:%7.3f(%7.3f) Mag:%7.3f norm:%6.3f PHASE[%3d][0]:%7.3f (%6.2f degrees)\n",
            Re, Re/F00, Im, Im/F00, mag, norm, u, phase, phase * 180.0 / PI );
       }
     }
#endif
   }
#ifdef TEST
 {
   float phase, mag, norm, F00;
   for(u=0;u<N;u++)
   {
     for(v=0;v<N;v++)
     {
       Re = Harmonics[u][v].Re;
       Im = Harmonics[u][v].Im;
       mag = (float) sqrt((double)(Re*Re+Im*Im));
       /* printf("%7.3f ", mag ); /**/
     }
     /* NL; /**/
   }
 }
#endif
}


void prepareFFT( TP, Harmonics )
Pixel **TP;
complex **Harmonics;
{
 int r, t;

 /* fill in the list with the polar points and initialize
        the 'Harmonics' for the FFT */
 NPolarPoints = 0;

 for( r = 0; r < N; r++ )
   for( t = 0; t < N; t++ )
   {
     Harmonics[r][t].Re = 0.0; Harmonics[r][t].Im = 0.0;

     if( TP[r][t] )
     {
       Harmonics[r][t].Re = (float)ON;

       /* Insert in list for FT */
       PolarPoints[NPolarPoints].r = r;
       PolarPoints[NPolarPoints].theta = t;
       NPolarPoints++;
     }
   }
   /* printf("NPolarPoints=%d", NPolarPoints ); DBG; /**/
}


static char buf[20];
/* for the NN input normalize between 0.0 and 1.0 */
#define SCALE_FACT (1.0)

static float scale_fact= SCALE_FACT;


void NormFFT( Harmonics, Amplitudes )
complex **Harmonics;
float **Amplitudes;
{
 float Ampl00; int u, v, i;

 Ampl00 = Amplitudes[0][0] * scale_fact;

 /* printf("\nNormFFT> Ampl00 = %f\n", Ampl00 );/**/
 for( u = 0; u < N; u++ )
   for( v = 0; v < NU; v++ )
   {
     Amplitudes[u][v] /= Ampl00;
     /* printf("\nNormFFT> Amplitudes[%d][%d]=%f\n",
	 u, v, Amplitudes[u][v] ); /**/
     /* CHECK */
     if(  Amplitudes[u][v] >  Ampl00 )
     {
       printf("Had Amplitudes[%d][%d]= %f bigger than Ampl[0][0] !!!\n",
                u,v, Amplitudes[u][v]); gets(buf); }
   }
}

