/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:54 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "def.h"
#include "unldef.h"

extern int		N;
extern int		NPolarPoints;
extern PolarPoint	*PolarPoints;

static float  *SinTab = NULL, *CosTab = NULL;


#ifdef PI
#undef PI
#endif
#define PI 3.14159265358979323846

void initFT()
{
 double arg;
 int i;

 SinTab = (float*) malloc( N * sizeof(float) ); CHKPTR(SinTab);
 CosTab = (float*) malloc( N * sizeof(float) ); CHKPTR(CosTab);

 for( i = 0; i < N; i++ )
 {
   arg = (double) (2.0 * PI * (float)i / (float)N);
   SinTab[i]= (float) sin( arg );
   CosTab[i]= (float) cos( arg );
 }
}


void closeFT()
{
  FREE( SinTab ); FREE( CosTab );
}


void calc_F00( F00 )
float *F00;
{
 int pp;
 float dc = 0.0;

 /* Set F[0][0] : sum all weighted polar pixels  */
 for( pp = 0; pp < NPolarPoints; pp++ )
   dc += (float)PolarPoints[pp].r;

 *F00 = dc / (float)N;
}


void calc_one_UFF( u, v, uff )
int u, v;	/* in */
float *uff;	/* out */
{
 int pp, IndS, rInd;
 float Re, Im;

 Re = 0.0;
 Im = 0.0;
 for( pp = 0; pp < NPolarPoints; pp++ )
 {
   IndS = (int)(u*PolarPoints[pp].r + v*PolarPoints[pp].theta) % N;
   rInd = (int)(PolarPoints[pp].r);
   Re += (float)rInd * CosTab[ IndS ];
   Im -= (float)rInd * SinTab[ IndS ];
 }
 Re = Re / (float)N; Im = Im / (float)N;
 *uff = (float)sqrt((double)(Re*Re+Im*Im));
}


/*
 * The routines calc_F00() and calc_one_UFF()
 * serve for the calculus of just a few UNL Fourier Features
 * for instance in the classification phase of a classifier
*/

void example_calc_UFF()
{
 int u, v;
 float F00, uff;

 calc_F00( &F00 );
 
 printf("uff[0][0] = %.5f\n", F00 );
 u=0;v=1; calc_one_UFF( u, v, &uff ); uff /= F00;
 printf(" uff[%3d][%3d] = %.5f\n", u, v, uff );
 u=3;v=2; calc_one_UFF( u, v, &uff ); uff /= F00;
 printf(" uff[%3d][%3d] = %.5f\n", u, v, uff );
 u=1;v=1; calc_one_UFF( u, v, &uff ); uff /= F00;
 printf(" uff[%3d][%3d] = %.5f\n", u, v, uff );
 u=3;v=1; calc_one_UFF( u, v, &uff ); uff /= F00;
 printf(" uff[%3d][%3d] = %.5f\n", u, v, uff );

}
