/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:58 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "def.h"

#define ALWAYS 1
#define OPTION 0

/*---------- P R O G R A M T Y P E ---------------------------------*/
#define PROGNAME1	"uff"		/* Identify type of program */
#define PROGNAME2	"fd"
/*------------------------------------------------------------------*/


#define RESOLUTION	"-N"		/* System resolution */
#define FRACTION	"-q"		/* Fraction of uff */
#define IN_DATA_DIR	"-dir"		/* name of input data directory */
#define UFF_OUT_FILE	"-uff"		/* Where to put the data */
#define FEAT_NAME	"-fnam"		/* name of the feature description */
#define VERBOSE		"-v"
#define OUT_ASCII	"-a"		/* Dump the values as text */
#define X_OFF		"-xoff"		/* deactivate the graphical output */
#define TRACE		"-t"		/* Wait after each image */
#define HELP1		"-h"
#define HELP2		"-help"


#define FD_OUT_FILE	"-fd"	/* Where to put the Fourier descriptors */
#define FD_BUF_LEN	"-res"	/* The resolution of the feature buffer*/
extern int FD_fraction;
extern int FD_buf_len;


extern int feature_type;
extern int UFF_fraction;
extern str100 dataDir;
extern bool is_2_power();
extern bool X;
extern bool trace;


str20 pr;				/* Prompt */
bool verbose = FALSE, help = FALSE;
int feature_type = EMPTY;
str100 featNameFile;
str100 featFile;
bool outAscii = FALSE;

int N = 128;


static str100 linebuf;
static str30 buf;
static char *num_str = NULL;
static char *in_data_dir = NULL;
static char *feat_out_file = NULL;
static char *sel_file = NULL;
static char *feat_name_file = NULL;
static bool *visited = NULL;  /* register all pararamters that were detected */


#define MAXLEN 300
static char *parse_input_param( argc, argv, param, when )
int argc;
char **argv;
char *param;
int when;
{
 int i = 0;
 char str[MAXLEN];

 while ((i < argc) && (strcmp(param, argv[i])))
   i++;
 if (i < argc-1) {
   visited[i] = TRUE;
   if( argv[i+1][0] != '-' )
     visited[i+1] = TRUE;
   else
   {
     fprintf( stderr,
	 "Could not find the argument for parameter %s - exit...\n", param );
     exit( 1 );
   }
   return(argv[i+1]);
 }
 else	/* Did not find parameter */
 {
   if (when == ALWAYS)
   {
     sprintf( str, "Can't find asked option %s", param);
     fprintf( stderr, "%s\n", str );
     exit(1);
   }
 }
 return((char *) NULL);
}

static bool find_input_param( argc, argv, param, when )
int argc;
char **argv;
char *param;
int when;
{
 char str[MAXLEN];
 bool found = FALSE;
 int i = 0;

 while( ! found && i < argc )
 {
   found = (strcmp(argv[i],param)==0);
   if( ! found )
     i++;
 }
 if (when == ALWAYS)
 {
   sprintf( str, "Can't find asked option %s", param);
   fprintf( stderr, "%s\n", str );
   exit(1);
 }
 if( found )
   visited[ i ] = TRUE;
 return( found );
} 

static void usage_UFF( argc, argv )
int argc;
char **argv;
{
 fprintf( stderr, "\nusage: %s [-h[elp]] [-N <Resolution>]  [-q <Fraction>]\n",
	 argv[0] );
 fprintf( stderr, "\t\t[-dir <data-directory>]  [-uff <uff-file>]\n" );
 fprintf( stderr, "\t\t[-v]  [-a]  [-t]  [-fnam <feature-names-file>]" );
#ifdef XWINDOW
 fprintf( stderr, "  [-xoff]" );
#endif
 fprintf( stderr, "\n\n" );
}


static void usage_FD( argc, argv )
int argc;
char **argv;
{
 fprintf( stderr, "\nusage: %s [-h[elp]] [-N <Image resolution>]", argv[0]);
 fprintf( stderr, "  [-q <Fraction>]\n" );
 fprintf( stderr, "\t\t[-res <Buffer resolution>]  [-dir <data-directory>]\n");
 fprintf( stderr, "\t\t[-fd <fd-file>]  [-fnam <feature-names-file>]\n" );
 fprintf( stderr, "\t\t[-v]  [-a]  [-t]" );
#ifdef XWINDOW
 fprintf( stderr, "  [-xoff]" );
#endif
 fprintf( stderr, "\n\n" );
}


static void parse_input_UFF( argc, argv )
int argc;
char **argv;
{
 int i;

 help = find_input_param( argc, argv, HELP1, OPTION ) ||
	find_input_param( argc, argv, HELP2, OPTION );
 if( help )
 {
   usage_UFF( argc, argv );
   exit(1);
 }
 verbose = find_input_param( argc, argv, VERBOSE, OPTION );
 outAscii = find_input_param( argc, argv, OUT_ASCII, OPTION );
 num_str = parse_input_param( argc, argv, RESOLUTION, OPTION );
 if( num_str != NULL )
 {
   sscanf( num_str, "%d", &i );
   if( ! is_2_power( i ) )
   {
     fprintf(stderr,"%s%d is not a power of 2 - using default %d...", pr, i, N);
     gets( buf );
   }
   else
     N = i; 
 }
 else
   fprintf(stderr,"%sUsing default image resolution %d\n", pr, N );


 num_str = parse_input_param( argc, argv, FRACTION, OPTION );
 if( num_str != NULL )
 {
   sscanf( num_str, "%d", &i );
   if( (!is_2_power( i ) && i!=1)  || i >= N )
   {
     fprintf(stderr,"%s%d invalid fraction - using default %d...",
	pr, i, UFF_fraction );
     gets( buf );
   }
   else
     UFF_fraction = i; 
 }
 else
   fprintf(stderr,"%sUsing default fraction %d\n", pr, UFF_fraction );


 in_data_dir = parse_input_param( argc, argv, IN_DATA_DIR, OPTION );
 dataDir[0] = '\0';
 if( in_data_dir != NULL )
 {
   strcpy( dataDir, in_data_dir );
 }
 else
 {
   linebuf[0] = '\0';
   do
   {
     printf("%sDirectory from where to read the PBM-images? ", pr );
     gets( linebuf );
   }
   while( linebuf[0] == '\0' );
   strcpy( dataDir, linebuf );
 }

 feat_out_file = parse_input_param( argc, argv, UFF_OUT_FILE, OPTION );
 if( feat_out_file != 0 )
 {
   strcpy( featFile, feat_out_file );
 }
 else
 {
   linebuf[0] = '\0';
   do
   {
     printf("%sFile to which write the UNL Fourier Features? ", pr);
     gets( linebuf );
   }
   while( linebuf[0] == '\0' );
   strcpy( featFile, linebuf );
 }


 feat_name_file = parse_input_param( argc, argv, FEAT_NAME, OPTION );
 if( feat_name_file != 0 )
 {
   strcpy( featNameFile, feat_name_file );
 }
 else
 {
   linebuf[0] = '\0';
   do
   {
     printf("%sOutput file with names of the UNL Fourier Features? ", pr);
     gets( linebuf );
   }
   while( linebuf[0] == '\0' );
   strcpy( featNameFile, linebuf );
 }
#ifdef XWINDOW
 X = ! find_input_param( argc, argv, X_OFF, OPTION );
#endif
 trace = find_input_param( argc, argv, TRACE, OPTION );
}


static void parse_input_FD( argc, argv )
int argc;
char **argv;
{
 int i;

 help = find_input_param( argc, argv, HELP1, OPTION ) ||
	find_input_param( argc, argv, HELP2, OPTION );
 if( help )
 {
   usage_FD( argc, argv );
   exit(1);
 }
 verbose = find_input_param( argc, argv, VERBOSE, OPTION );
 outAscii = find_input_param( argc, argv, OUT_ASCII, OPTION );
 num_str = parse_input_param( argc, argv, RESOLUTION, OPTION );
 if( num_str != NULL )
 {
   sscanf( num_str, "%d", &i );
   if( ! is_2_power( i ) )
   {
     fprintf(stderr,"%s%d is not a power of 2 - using default %d...",
		pr, i, N );
     gets( buf );
   }
   else
     N = i; 
 }
 else
   fprintf(stderr,"%sUsing default image resolution %d\n", pr, N );


 num_str = parse_input_param( argc, argv, FD_BUF_LEN, OPTION );
 if( num_str != NULL )
 {
   sscanf( num_str, "%d", &i );
   if( (!is_2_power( i ) && i!=1) )
   {
     fprintf(stderr,"%s%d invalid resolution of buffer - using default %d...",
	pr, i, FD_buf_len );
     gets( buf );
   }
   else
     FD_buf_len = i; 
 }
 else
   fprintf(stderr,"%sUsing default buffer resolution %d\n", pr, FD_buf_len );


 num_str = parse_input_param( argc, argv, FRACTION, OPTION );
 if( num_str != NULL )
 {
   sscanf( num_str, "%d", &i );
   if( (!is_2_power( i ) && i!=1)  || i >= FD_buf_len )
   {
     fprintf(stderr,"%s%d invalid fraction - using default %d...",
	pr, i, FD_fraction );
     gets( buf );
   }
   else
     FD_fraction = i; 
 }
 else
   fprintf(stderr,"%sUsing default fraction %d\n", pr, FD_fraction );


 in_data_dir = parse_input_param( argc, argv, IN_DATA_DIR, OPTION );
 dataDir[0] = '\0';
 if( in_data_dir != NULL )
 {
   strcpy( dataDir, in_data_dir );
 }
 else
 {
   linebuf[0] = '\0';
   do
   {
     printf("%sDirectory from where to read the PBM-images? ", pr );
     gets( linebuf );
   }
   while( linebuf[0] == '\0' );
   strcpy( dataDir, linebuf );
 }

 feat_out_file = parse_input_param( argc, argv, FD_OUT_FILE, OPTION );
 if( feat_out_file != 0 )
 {
   strcpy( featFile, feat_out_file );
 }
 else
 {
   linebuf[0] = '\0';
   do
   {
     printf("%sWrite the Fourier Descriptors to file? ", pr );
     gets( linebuf );
   }
   while( linebuf[0] == '\0' );
   strcpy( featFile, linebuf );
 }


 feat_name_file = parse_input_param( argc, argv, FEAT_NAME, OPTION );
 if( feat_name_file != 0 )
 {
   strcpy( featNameFile, feat_name_file );
 }
 else
 {
   linebuf[0] = '\0';
   do
   {
     printf("%sWrite the names of the Fourier Descriptors to file? ", pr );
     gets( linebuf );
   }
   while( linebuf[0] == '\0' );
   strcpy( featNameFile, linebuf );
 }
#ifdef XWINDOW
 X = ! find_input_param( argc, argv, X_OFF, OPTION );
#endif
 trace = find_input_param( argc, argv, TRACE, OPTION );
}


void parse_input( argc, argv )
int argc;
char **argv;
{
 int i;

 /* detect the program */
 if( (strcmp(PROGNAME1,argv[0])==0) )
 {
   feature_type = UFF;
   strcpy( pr, "UFF$ ");
 }
 if( (strcmp(PROGNAME2,argv[0])==0) )
 {
   feature_type = FOURIER_DESCRIPTORS;
   strcpy( pr, "FD$ ");
 }

 visited = (bool*) malloc( argc*sizeof(bool) );
 for( i = 0; i < argc; i++ )
   visited[i] = FALSE;
 visited[ 0 ] = TRUE;	/* argv[0] */

 switch( feature_type )
 {
   case UFF: parse_input_UFF( argc, argv ); break;
   case FOURIER_DESCRIPTORS: parse_input_FD( argc, argv ); break;
   default:
     fprintf(stderr,"Could not identify the program name \"%s\". Exit...\n",
		argv[0]);
     fprintf(stderr,"Should be \"uff\" or \"fd\"\n"); exit(1);
 }

 /* check if all parameters and arguments were visited */
 for( i = 0; i < argc; i++ )
   if( ! visited[i] )
   {
     fprintf(stderr,
	"%sUnknown option \"%s\" or missing parameter - Exit...\n",pr, argv[i]);
     FREE( visited );
     exit(1);
   }
 FREE( visited );
}
