/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:56 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <string.h>

#ifdef XWINDOW
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define MIN(a,b)               (a>b ? b : a)
#define MAX(a,b)               (a<b ? b : a)

#define TOL 150
#define TOPMARGIN 200
#define MIN_WIDTH 350
#define MIN_HEIGHT 200

extern int N;

static int win_width, winheight;
static char winTitle[22] = "2-D Shape Description";

static Display *tgi_display = NULL;
static Window  tgi_window;
static GC tgi_gc;
static KeySym  tgi_key;
static XSizeHints tgi_hint;
static int tgi_screen;
static unsigned long tgi_foreground, tgi_background;


void tgi_ClearArea(x,y,width,height)
 int x,y,width,height;
{ XClearArea(tgi_display,tgi_window,x,y,width,height,0); }


void tgi_drawlines(Npts,Points)
 int Npts; short int *Points;
{ XDrawLines(tgi_display,tgi_window,tgi_gc,
                      (XPoint *)Points,Npts,CoordModeOrigin); }


void tgi_init()
{
  int argc=0;
  char **argv=NULL;

  char *tgi_display_name = (char*)0;

  /* init */
  tgi_display = XOpenDisplay( tgi_display_name );
  if( tgi_display == NULL )
  {
    fprintf( stderr, "Could not open display. Exit...\n" );
    exit(1);
  }
  tgi_screen = DefaultScreen(tgi_display);
  
  tgi_background = WhitePixel(tgi_display, tgi_screen);
  tgi_foreground = BlackPixel(tgi_display, tgi_screen);

  tgi_hint.x = TOL; tgi_hint.y = TOL;
  tgi_hint.width = MAX(2*N+TOL,MIN_WIDTH);
  tgi_hint.height = MAX(N+TOPMARGIN,MIN_HEIGHT);
  tgi_hint.flags = PPosition | PSize;

  tgi_window = XCreateSimpleWindow(tgi_display,
     DefaultRootWindow(tgi_display),
     tgi_hint.x, tgi_hint.y, tgi_hint.width, tgi_hint.height,
     5, tgi_foreground, tgi_background);
  XSetStandardProperties(tgi_display, tgi_window, winTitle, winTitle,
     None, argv, argc, &tgi_hint);

  tgi_gc = XCreateGC(tgi_display, tgi_window, 0, 0);
  XSetBackground( tgi_display, tgi_gc, tgi_background);
  XSetForeground( tgi_display, tgi_gc, tgi_foreground);
  XMapRaised(tgi_display, tgi_window);
  XFlush(tgi_display);
}

void tgi_close()
{
  XFreeGC(tgi_display,tgi_gc);
  XDestroyWindow(tgi_display,tgi_window);
  XFlush(tgi_display);
  XCloseDisplay(tgi_display);
}

void tgi_pixel(x,y)
int x,y;
{ XDrawPoint(tgi_display,tgi_window,tgi_gc,x,y); }

void tgi_flush()
{ XFlush(tgi_display); }

void tgi_clearWin()
{ XClearWindow(tgi_display,tgi_window); }

void tgi_line(x1,y1,x2,y2)
int x1,y1,x2,y2;
{ XDrawLine(tgi_display,tgi_window,tgi_gc,x1,y1,x2,y2); }

void tgi_rectangle(x,y,wd,ht)
int x,y,wd,ht;
{ XDrawRectangle(tgi_display,tgi_window,tgi_gc,x,y,wd,ht); }

void tgi_string(x,y,str)
int x,y; char *str;
{ XDrawString(tgi_display,tgi_window,tgi_gc,x,y,str,strlen(str)); }

#else
/* dummy functions without Xwindow */
void tgi_ClearArea(_1,_2,_3,_4) int _1,_2,_3,_4; {};
void tgi_drawlines(_1,_2) int _1; short int _2; {};
void tgi_init(){};
void tgi_close(){};
void tgi_pixel(_1,_2) int _1,_2; {};
void tgi_flush(){};
void tgi_clearWin(){};
void tgi_line(_1,_2,_3,_4) int _1,_2,_3,_4; {};
void tgi_rectangle(_1,_2,_3,_4) int _1,_2,_3,_4; {};
void tgi_string(_1,_2,_3) int _1, _2; char *_3; {};
#endif
