/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:52 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "def.h"
#include "unldef.h"

extern int N;

static char x(i)
int i;
{ if( i ) return('*'); else return(' '); }

static int peepMask( p1, p2, p3, p4, p5, p6, p7, p8, p9 )
int p1, p2, p3, p4, p5, p6, p7, p8, p9;
{ printf(" _____\n");
  printf(" |%c%c%c|\n",x(p9),x(p2),x(p3));
  printf(" |%c%c%c|\n",x(p8),x(p1),x(p4));
  printf(" |%c%c%c|\n",x(p7),x(p6),x(p5));
  printf(" -----\n");
}

/*>>>>>>>>>>-------------------------------------<<<<<<<<<<<<<<<*/
/*								*/
/*     		  Thinning algorithm by Zhang & Suen		*/
/*	c.f. Gonzalez / Wintz : Digital Image Processing	*/
/*          II. edition, page 399, Addison Wesley		*/
/*								*/
/*>>>>>>>>>>--------------------------------------<<<<<<<<<<<<<<*/

/* use Lookup-tables to look up thinning condition */
static Pixel ThinLookupStep1[2][2][2][2][2][2][2][2][2];
static Pixel ThinLookupStep2[2][2][2][2][2][2][2][2][2];


static int getCondB( p2, p3, p4, p5, p6, p7, p8, p9 )
int p2, p3, p4, p5, p6, p7, p8, p9;
{ int CondB, Sp1= 0; 
 
  if( !p2 && p3 ) Sp1++;
  if( !p3 && p4 ) Sp1++;
  if( Sp1 > 1 ) CondB= FALSE;
  else
  {
    if( !p4 && p5 ) Sp1++;
    if( Sp1 > 1 ) CondB= FALSE;
    else
    {
      if( !p5 && p6 ) Sp1++;
      if( Sp1 > 1 ) CondB= FALSE;
      else
      {
        if( !p6 && p7 ) Sp1++;
        if( Sp1 > 1 ) CondB= FALSE;
        else
        {
          if( !p7 && p8 ) Sp1++;
          if( Sp1 > 1 ) CondB= FALSE;
          else
          {
            if( !p8 && p9 ) Sp1++;
            if( Sp1 > 1 ) CondB= FALSE;
            else
            {
              if( !p9 && p2 ) Sp1++;
              if( Sp1 > 1 ) CondB= FALSE;
              else CondB= Sp1== 1;
            }
          }
        }
      }
    }
  }
  return( CondB ); 
}


void initThinLookup()
{ int p1, p2, p3, p4, p5, p6, p7, p8, p9;
 int Np1, CondA, CondB, CondC1, CondD1, CondC2, CondD2, Cond1, Cond2;

 for( p1= 0; p1< 2; p1++ )
   for( p2= 0; p2< 2; p2++ )
     for( p3= 0; p3< 2; p3++ )
       for( p4= 0; p4< 2; p4++ )
         for( p5= 0; p5< 2; p5++ )
           for( p6= 0; p6< 2; p6++ )
             for( p7= 0; p7< 2; p7++ )
               for( p8= 0; p8< 2; p8++ )
                 for( p9= 0; p9< 2; p9++ ) {
 
                   Np1= p2+p3+p4+p5+p6+p7+p8+p9;
                   CondA= ( (2 <= Np1) && (Np1 <= 6) && p1 );
                   if( CondA ) {
                     CondB= getCondB(p2, p3, p4, p5, p6, p7, p8, p9);
                     if( CondB ) {

                       /* STEP 1 */
                       CondC1= !(p2 && p4 && p6);
                       if( CondC1 ) {
                         CondD1= !(p4 && p6 && p8);
                         Cond1= CondD1;
                       }
                       else Cond1= FALSE;

                       /* STEP 2 */
                       CondC2= !(p2 && p4 && p8);
                       if( CondC2 ) {
                         CondD2= !(p2 && p6 && p8);
                         Cond2= CondD2;
                       }
                       else Cond2= FALSE;
                       /* end STEP2 */

                     }
                     else { Cond1= FALSE; Cond2= FALSE; }
                   }
                   else { Cond1= FALSE; Cond2= FALSE; }
                   /* Now it is determined if condition 1 and 2 are o.k. */
                   /* pixel will stay ON if center pixel is present and
                      Condition is not satisfied; else OFF */

                   ThinLookupStep1[p1][p2][p3][p4][p5][p6][p7][p8][p9]=
                       (Pixel)( p1 && !Cond1 );

                   ThinLookupStep2[p1][p2][p3][p4][p5][p6][p7][p8][p9]=
                       (Pixel)( p1 && !Cond2 );
                 }
}


extern Pixel **alloc_pattern();

/* 
 * NORMAL CASE : PATTERN IN CARTESIAN COORDINATES 
*/
void thin_Zhang_Suen( P )
Pixel **P;
{
 Pixel **Q = NULL, **R = NULL;
 int x, y;
 int xMin, xMax, yMin, yMax;
 int nrPointsDeleted, delStep1, delStep2, nrPasses=0, Np1, skeleton= FALSE;

 Q = alloc_pattern( N ); CHKPTR( Q );
 R = alloc_pattern( N ); CHKPTR( R );;

 /* calculate the bounding box of the binary pattern */ 
 xMin= N-1; xMax= 0; yMin= N-1; yMax= 0;
 for( y=0; y< N; y++ )
   for( x=0; x< N; x++ )
     if( P[x][y] )
     {
       if( x < xMin ) xMin= x;
       if( y < yMin ) yMin= y;
       if( x > xMax ) xMax= x;
       if( y > yMax ) yMax= y;
     }
 if( xMin < 2 ) xMin= 2; if( yMin < 2 ) yMin= 2;
 if( xMax > N-3 ) xMax= N-3; if( yMax > N-3 ) yMax= N-3;

 while( !skeleton )
 {
  clearPattern( Q ); clearPattern( R );
  nrPointsDeleted= 0; delStep1= 0; delStep2= 0;

  /*** STEP 1 ***/
  /* scan the pattern with the 3x3 mask (8-neighborhood) */
  for(y= yMin-2; y<= yMax; y++)
    for(x= xMin-2; x<= xMax; x++)
    {

      Q[x+1][y+1]=
          ThinLookupStep1[P[x+1][y+1]][P[x+1][y+2]][P[x+2][y+2]][P[x+2][y+1]][P[x+2][y]][P[x+1][y]][P[x][y]][P[x][y+1]][P[x][y+2]];

      if( P[x+1][y+1] != Q[x+1][y+1] ) delStep1++;
/*
if( P[x+1][y+1] != Q[x+1][y+1] ) {
  peepMask( P[x+1][y+1], P[x+1][y+2], P[x+2][y+2], P[x+2][y+1], P[x+2][y], P[x+1][y], P[x][y], P[x][y+1], P[x][y+2] );
  printf("THIN> ---> Condition 111 is satisfied. Point will be del.\n"); } /* */

    }/* for */

  /*** STEP 2 ***/
  for(y= yMin-2; y<= yMax; y++)
    for(x= xMin-2; x<= xMax; x++)
    {

     R[x+1][y+1]=
          ThinLookupStep2[Q[x+1][y+1]][Q[x+1][y+2]][Q[x+2][y+2]][Q[x+2][y+1]][Q[x+2][y]][Q[x+1][y]][Q[x][y]][Q[x][y+1]][Q[x][y+2]];

      if( Q[x+1][y+1] != R[x+1][y+1] ) delStep2++;
/*
if( Q[x+1][y+1] != R[x+1][y+1] ) {
  peepMask( Q[x+1][y+1], Q[x+1][y+2], Q[x+2][y+2], Q[x+2][y+1], Q[x+2][y], Q[x+1][y], Q[x][y], Q[x][y+1], Q[x][y+2] );
  printf("THIN> ---> Condition 222 is satisfied. Point will be del.\n"); } /* */

    }/* for */

  copyPattern( R, P );
  nrPointsDeleted= delStep1 + delStep2;
  skeleton= (nrPointsDeleted == 0); 
  nrPasses++;
/* 
  printf("Thin__Zhang_Suen> PASS: %d , Number of deleted points: Step1: %d",
          nrPasses, delStep1);
  printf(" + Step2: %d = %d\n",delStep2, nrPointsDeleted); /* */
 }/*while*/  

 free_pattern( Q, N ); free_pattern( R, N );
}


/* 
 * SPECIAL CASE : PATTERN IN POLAR COORDINATES
 *  close the loop between smallest and biggest angle 
*/
void thin_Zhang_Suen_Polar( P )
Pixel **P;
{
 Pixel **Q = NULL, **R = NULL;
 int x, y;
 int nrPointsDeleted, delStep1, delStep2, nrPasses=0, Np1, skeleton= FALSE;
 Pixel p1, p2, p3, p4, p5, p6, p7, p8, p9;
 Pixel *MaxAngle = NULL, *MinAngle = NULL;

 MaxAngle = (Pixel*) malloc( N * sizeof(Pixel) ); CHKPTR( MaxAngle );
 MinAngle = (Pixel*) malloc( N * sizeof(Pixel) ); CHKPTR( MinAngle );

 Q = alloc_pattern( N ); CHKPTR( Q );
 R = alloc_pattern( N ); CHKPTR( R );;

 while( !skeleton )
 {
  clearPattern( Q ); clearPattern( R );
  nrPointsDeleted= 0; delStep1= 0; delStep2= 0;

  /* close the loop: wrap pattern around theta */
  for( x= 0; x< N; x++ ) MinAngle[x] = P[x][N-1];
  for( x= 0; x< N; x++ ) MaxAngle[x] = P[x][0];

  /*>>>-------<<<*/
  /*** STEP 1  ***/
  /*>>>-------<<<*/
  /* scan the pattern with the 3x3 mask (8-neighborhood) */

  /* Special Case : first row */

  for( x = 1; x< N-1; x++ )
  {
    Q[x][0]=
        ThinLookupStep1[P[x][0]][P[x][1]][P[x+1][1]][P[x+1][0]][MinAngle[x+1]][MinAngle[x]][MinAngle[x-1]][P[x-1][0]][P[x-1][1]];
    if( P[x][0] != Q[x][0] ) delStep1++;
  }

  /* Last row */

  for( x= 1; x< N-1; x++ )
  {
    Q[x][N-1]=
        ThinLookupStep1[P[x][N-1]][MaxAngle[x]][MaxAngle[x+1]][P[x+1][N-1]][P[x+1][N-2]][P[x][N-2]][P[x-1][N-2]][P[x-1][N-1]][MaxAngle[x-1]];
    if( P[x][N-1] != Q[x][N-1] ) delStep1++;
  }

  /* rest */
  for(y= 0; y< N-2; y++)
    for(x= 0; x< N-2; x++)
    {

      Q[x+1][y+1]=
          ThinLookupStep1[P[x+1][y+1]][P[x+1][y+2]][P[x+2][y+2]][P[x+2][y+1]][P[x+2][y]][P[x+1][y]][P[x][y]][P[x][y+1]][P[x][y+2]];

      if( P[x+1][y+1] != Q[x+1][y+1] ) delStep1++;
/*
if( P[x+1][y+1] != Q[x+1][y+1] ) {
  peepMask( P[x+1][y+1], P[x+1][y+2], P[x+2][y+2], P[x+2][y+1], P[x+2][y], P[x+1][y], P[x][y], P[x][y+1], P[x][y+2] );
  printf("THIN> ---> Condition 111 is satisfied. Point will be del.\n"); } /* */

    }/* for */

  /* 
   *  copy the first and the last column and process them
   *
  */
  for( y= 0; y< N; y++ )
  { 
    Q[0][y]= P[0][y]; Q[N-1][y]= P[N-1][y];
  }
  /* lower left corner */
  Q[0][0]=
       ThinLookupStep1[P[0][0]][P[0][1]][P[1][1]][P[1][0]][OFF][OFF][OFF][OFF][OFF];
  if( P[0][0] != Q[0][0] ) delStep1++;

  /* upper left corner */
  Q[0][N-1]=
       ThinLookupStep1[P[0][N-1]][OFF][OFF][P[1][N-1]][P[1][N-2]][P[0][N-2]][OFF][OFF][OFF];
  if( P[0][N-1] != Q[0][N-1] ) delStep1++;

  /* rest of first column */
  for( y= 1; y< N-2; y++ )
  {
    Q[0][y]=
       ThinLookupStep1[P[0][y]][P[0][y+1]][P[1][y+1]][P[1][y]][P[1][y-1]][P[0][y-1]][OFF][OFF][OFF];
    if( P[0][y] != Q[0][y] ) delStep1++;
  }

  /* lower right corner */
  Q[N-1][0]=
       ThinLookupStep1[P[N-1][0]][P[N-1][1]][OFF][OFF][OFF][OFF][OFF][P[N-2][0]][P[N-2][1]];
  if( P[N-1][0] != Q[N-1][0] ) delStep1++;

  /* upper right corner */
  Q[N-1][N-1]=
       ThinLookupStep1[P[N-1][N-1]][OFF][OFF][OFF][OFF][P[N-1][N-2]][P[N-2][N-2]][P[N-2][N-1]][OFF];
  if( P[N-1][N-1] != Q[N-1][N-1] ) delStep1++;

  /* rest of last column */
  for( y= 1; y< N-2; y++ )
  {
    Q[N-1][y]=
       ThinLookupStep1[P[N-1][y]][P[N-1][y+1]][OFF][OFF][OFF][P[N-1][y-1]][P[N-2][y-1]][P[N-2][y]][P[N-2][y+1]];
    if( P[N-1][y] != Q[N-1][y] ) delStep1++;
  }


  /*>>>-------<<<*/
  /*** STEP 2  ***/
  /*>>>-------<<<*/

  /* close the loop : wrap pattern around theta */
  for( x= 0; x< N; x++ ) MinAngle[x] = Q[x][N-1];
  for( x= 0; x< N; x++ ) MaxAngle[x] = Q[x][0];

  /* Special Case : first row */

  for( x = 1; x< N-1; x++ )
  {
    R[x][0]=
        ThinLookupStep2[Q[x][0]][Q[x][1]][Q[x+1][1]][Q[x+1][0]][MinAngle[x+1]][MinAngle[x]][MinAngle[x-1]][Q[x-1][0]][Q[x-1][1]];
    if( Q[x][0] != R[x][0] ) delStep2++;
  }

  /* Last row */

  for( x= 1; x< N-1; x++ )
  {
    R[x][N-1]=
        ThinLookupStep2[Q[x][N-1]][MaxAngle[x]][MaxAngle[x+1]][Q[x+1][N-1]][Q[x+1][N-2]][Q[x][N-2]][Q[x-1][N-2]][Q[x-1][N-1]][MaxAngle[x-1]];
    if( Q[x][N-1] != R[x][N-1] ) delStep2++;
  }

  /* rest */
  for(y= 0; y< N-2; y++)
    for(x= 0; x< N-2; x++)
    {
      R[x+1][y+1]=
          ThinLookupStep2[Q[x+1][y+1]][Q[x+1][y+2]][Q[x+2][y+2]][Q[x+2][y+1]][Q[x+2][y]][Q[x+1][y]][Q[x][y]][Q[x][y+1]][Q[x][y+2]];

      if( Q[x+1][y+1] != R[x+1][y+1] ) delStep2++;
/*
if( Q[x+1][y+1] != R[x+1][y+1] ) {
  peepMask( Q[x+1][y+1], Q[x+1][y+2], Q[x+2][y+2], Q[x+2][y+1], Q[x+2][y], Q[x+1][y], Q[x][y], Q[x][y+1], Q[x][y+2] );
  printf("THIN> ---> Condition 222 is satisfied. Point will be del.\n"); } /* */

    }/* for */


  /* 
   *  copy the first and the last column and process them
   *
  */
  for( y= 0; y< N; y++ )
  { 
    R[0][y]= Q[0][y]; R[N-1][y]= Q[N-1][y];
  }

  /* lower left corner */
  R[0][0]=
       ThinLookupStep1[Q[0][0]][Q[0][1]][Q[1][1]][Q[1][0]][OFF][OFF][OFF][OFF][OFF];
  if( Q[0][0] != R[0][0] ) delStep2++;

  /* upper left corner */
  R[0][N-1]=
       ThinLookupStep1[Q[0][N-1]][OFF][OFF][Q[1][N-1]][Q[1][N-2]][Q[0][N-2]][OFF][OFF][OFF];
  if( Q[0][N-1] != R[0][N-1] ) delStep2++;

  /* rest of first column */
  for( y= 1; y< N-2; y++ )
  {
    R[0][y]=
       ThinLookupStep1[Q[0][y]][Q[0][y+1]][Q[1][y+1]][Q[1][y]][Q[1][y-1]][Q[0][y-1]][OFF][OFF][OFF];
    if( Q[0][y] != R[0][y] ) delStep2++;
  }

  /* lower right corner */
  R[N-1][0]=
       ThinLookupStep1[Q[N-1][0]][Q[N-1][1]][OFF][OFF][OFF][OFF][OFF][Q[N-2][0]][Q[N-2][1]];
  if( Q[N-1][0] != R[N-1][0] ) delStep2++;

  /* upper right corner */
  R[N-1][N-1]=
       ThinLookupStep1[Q[N-1][N-1]][OFF][OFF][OFF][OFF][Q[N-1][N-2]][Q[N-2][N-2]][Q[N-2][N-1]][OFF];
  if( Q[N-1][N-1] != R[N-1][N-1] ) delStep2++;

  /* rest of last column */
  for( y= 1; y< N-2; y++ )
  {
    R[N-1][y]=
       ThinLookupStep1[Q[N-1][y]][Q[N-1][y+1]][OFF][OFF][OFF][Q[N-1][y-1]][Q[N-2][y-1]][Q[N-2][y]][Q[N-2][y+1]];
    if( Q[N-1][y] != R[N-1][y] ) delStep2++;
  }


  copyPattern( R, P );
  nrPointsDeleted = delStep1 + delStep2;
  skeleton = (nrPointsDeleted == 0); 
  nrPasses++;
/*
  printf("Thin__Zhang_Suen_Polar> PASS: %d ,
        Number of deleted points: Step1: %d", nrPasses, delStep1);
  printf(" + Step2: %d = %d\n",delStep2, nrPointsDeleted); /* */
 }/*while*/  

 free_pattern( Q, N ); free_pattern( R, N );
 FREE( MinAngle ); FREE( MaxAngle );
}
