/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:55 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include "def.h"
#include "unldef.h"

extern str20 pr;
extern bool verbose;
extern bool X;
extern int feature_type;
extern bool outAscii;
extern int N;
extern str100 featNameFile;
extern FILE *featData;
extern int nrFeat;


int NU;
int UFF_fraction = 8;

Pixel		**TP = NULL;
int		NPolarPoints;
PolarPoint	*PolarPoints = NULL;
complex		**Harmonics = NULL;
float		**Amplitudes = NULL;


static str80 buf;



void Alloc_UNL_Trans()
{
 int i;

 Alloc_Orig_Pat();

 Harmonics = (complex**) malloc( N * sizeof(complex*) );
 for( i = 0; i < N; i++ )
   Harmonics[i] = (complex*) malloc( N * sizeof(complex) );
 CHKPTR(Harmonics);

 Amplitudes = (float**) malloc( N * sizeof(float*) );
 for( i = 0; i < N; i++ )
   Amplitudes[i] = (float*) malloc( NU * sizeof(float) );
 CHKPTR(Amplitudes);
 
 PolarPoints = (PolarPoint*) malloc( N*N*sizeof(struct PolarPoint_));
 CHKPTR(PolarPoints);
}


void Free_UNL_Trans()
{
 int i;

 Free_Orig_Pat();

 for( i = 0; i < N; i++ )
   FREE(Harmonics[i]);
 FREE(Harmonics);

 for( i = 0; i < N; i++ )
   FREE(Amplitudes[i]);
 FREE(Amplitudes);
 
 FREE(PolarPoints);
}


/* IMPORTANT: consider conjugate symmetry and ignore F(0,0) */
static bool conjugate_sym_condition( u, v )
int u, v;
{
 return( ! ( (u==0 && v==0) || (u>NU-1 && v==0) || (u>NU-1 && v==NU-1)  ) );
}


/*
 * Limit the UNL Fourier Features only to the lower frequencies, i. e.
 * the v-parameter is small and
 * the u-parameter is small or near its maximum value
*/
static bool bounding_condition( u, v )
int u, v;
{
 bool valid_uff; int uabs;

 valid_uff = ( v <= NU / UFF_fraction );
 if( u  >= N/2 )        /* upper half */
   uabs = N-u;
 else                   /* lower half */
   uabs = u;
 valid_uff = valid_uff && ( uabs <= N / UFF_fraction );
 return( valid_uff );
}


void gen_uff_feat_desc()
{
 FILE *fd = NULL;
 int i = 0, u, v, uchr, vchr;

 fd = fopen( featNameFile, f_open_text_w );
 if( fd == NULL )
    { fprintf(stderr,"%sCould not open %s for writing !!!\n",
	pr, featNameFile ); exit(1); }
 printf("%sDumping feature name description to %s\n", pr, featNameFile  );

 for( u = 0; u < N; u++ )
   for( v = 0; v < NU; v++ )
   {
     if( conjugate_sym_condition( u, v ) &&  bounding_condition( u, v ) )
       i++;
   }
 nrFeat = i;
 fprintf( fd, "%d\n", nrFeat );

 for( u = 0; u < N; u++ )
   for( v = 0; v < NU; v++ )
   {
     if( conjugate_sym_condition( u, v ) &&  bounding_condition( u, v ) )
     {
       vchr = v;
       if( u >= 0 && u <= N/2 )
         uchr = u;
       else
         uchr = u - N;
       fprintf( fd, "u%d_v%d\n", uchr, vchr );
     }
   }
 fclose( fd );
}


#define NEXTLINE 20
void dumpFeat_UFF()
{
 int u, v, nr = 1;
 float uff;

 for( u = 0; u < N; u++ )
   for( v = 0; v < NU; v++ )
   {
     uff = Amplitudes[u][v];
     if( conjugate_sym_condition( u, v ) &&  bounding_condition( u, v ) )
     {
       /* printf("uff[%3d][%3d]=%0.5f \n", u, v, uff ); /**/
       if( outAscii )
       {
         fprintf( featData, "%f ", uff );
         if( (nr % NEXTLINE) == 0 )
           fprintf( featData, "\n");
         nr++;
       }
       else
         fwrite( &uff, sizeof(float), 1, featData );
     }
   }
}
