/************************************************************************
 *                                                                      *
 *  Program package         U F F                                       *
 *                                                                      *
 *  Version   1.2                                                       *
 *  Date:     Fri Nov  1 11:11:57 1996                                  *
 *                                                                      *
 *  NOTE: This program package is copyrighted in the sense that it      *
 *  may be used for scientific purposes. The package as a whole, or     *
 *  parts thereof, cannot be included or used in any commercial         *
 *  application without written permission granted by the author.       *
 *  No programs contained in this package may be copied for commercial  *
 *  distribution.                                                       *
 *                                                                      *
 *  All comments  concerning this program package may be sent to the    *
 *  e-mail address 'tr@uninova.pt'.                                     *
 *                                                                      *
 ************************************************************************/



#include <stdio.h>
#include <math.h>
#include "def.h"
#include "unldef.h"

#ifdef PI
#undef PI
#endif
#define PI 3.14159265358979323846
#ifdef PI2
#undef PI2
#endif
#define PI2 6.2831853071795862320

extern int N;
extern int NPoints;
extern CPoint *Points;

static float maxDist;
static CartesianCoord O;


/*
 * Euclidean distance between two points
*/
static void EuclidDist( P1, P2, D )
CPoint *P1;
CartesianCoord *P2;
float *D;
{
 float x, y;

 x = P1->x - P2->x;
 y = P1->y - P2->y;
 *D = sqrt( x*x + y*y );
}


/*
 * Discrete centroid of several points
*/
static void calcCentroid( P )
Pixel **P;
{
 int i, x, y;
 float dist;

 O.x = 0.0; O.y = 0.0; NPoints = 0;
 for( x = 0; x < N; x++ )
   for( y = 0; y < N; y++ )
   {
     if ( P[x][y] )
     {
       O.x = O.x + x; O.y = O.y + y;
       Points[NPoints].x = x;  Points[NPoints].y = y;
       NPoints += 1;
     }
   }
 /* printf("Number of Points in original Pattern: %d", NPoints); DBG; /**/
 O.x /= (float)NPoints;
 O.y /= (float)NPoints;
 maxDist = 0.0;
 for ( i = 0; i < NPoints; i++ )
 {
   EuclidDist(&(Points[i]), &O, &dist);
   if ( dist > maxDist)
     maxDist = dist;
 }
}


/*** Transforms from the Cartesian to the polar coordinate system */
static void coordTransform( x, y, R, t)
float x, y;
int *R, *t;
{
 double r, theta, dx,dy;
 
 dx = (double)(x - O.x); dy = (double)(y - O.y);
#ifdef NON_UNIX
 if( dy == 0.0 || dx == 0.0 )
   theta = 0.0;
 else
   theta = atan2( dy, dx );
#else
 theta = atan2( dy, dx );
#endif
 r = sqrt( dx*dx + dy*dy );

 *R= ROUND( (float)(N-1)*r/maxDist );
 if( theta >= 0 )
   *t = ROUND( (float)(N-1)*theta/PI2 );
 else
   *t = ROUND( (float)(N-1)*(PI2+theta)/PI2 );
}


/*** transforming from Cartesian to polar by representing line in 
*** parameter form and generating curve by this parameter 
***/
static void intpolPolarCurveOld( x1, y1, x2, y2, P )
int x1, y1, x2, y2;
Pixel **P;
{
 int dx, dy, i, Steps, R, t, R1, R2, t1, t2;
 int dR, dt;
 float x, y, di, dp;

 /* get the two endpoints in the polar coordinates */

 coordTransform( (float)x1, (float)y1, &R1, &t1 );
 P[R1][t1] = ON;
 coordTransform( (float)x2, (float)y2, &R2, &t2 );
 P[R2][t2] = ON;

 /* calculate the number of steps necessary to interpolate
    without any holes.
    Choose the Euclidean distance between the two polar points
    (R1,t1) and (R2,t2) and multiply with a constant, say 2*Pi
    Comment: the exact number of steps has still to be explored
 */

 dR = R2 - R1; dt = t2 - t1;
 dp = (float)sqrt( (double)(dR*dR + dt*dt) );
 Steps = 1 + ROUND( PI2 * dp );

 /* if( R1 < 10 && R2 < 10 ) {
   printf("R1: %d R2: %d t1: %d t2: %d\n",R1,R2,t1,t2);
   printf("Steps: %d\n",Steps); } /**/

 di= (float)Steps;
 
 dx= x2-x1; dy = y2-y1;

 for( i = 1; i < Steps; i++)
 {
   x = (float)(x1 + i/di * dx);
   y = (float)(y1 + i/di * dy);

   coordTransform( x, y, &R, &t );
   P[R][t] = ON;
 }
}


/*** 
*** Test if two pixels are 8-neighbors
***/
static bool neib8( Ra, ta, Rb, tb )
int Ra, ta, Rb, tb;
{
  return( ( ABS(Ra - Rb)%(N-1) <= 1 ) && ( ABS(ta - tb)%(N-1) <= 1 ) );
}


/*** 
*** Recursivly fill the transformed curve in the polar coordinate system
***/
static void fillPolarCurve( xl, yl, xr, yr, Rl, tl, Rr, tr, P )
float xl, yl, xr, yr;
int Rl, tl, Rr, tr;
Pixel **P;
{
 float xm, ym;
 int Rm, tm;

 /* printf("(xl,yl)=(%15.12f,%15.12f) (xr,yr)=(%15.12f,%15.12f) - ",
	xl, yl, xr, yr );
 printf("(Rl,tl)=(%3d,%3d) (Rr,tr)=(%3d,%3d)", Rl, tl, Rr, tr ); NL;	/**/

 if( ! (neib8( Rl, tl, Rr, tr ) || (Rl==0 && Rr==0)) )
 {
   xm = xl + (xr-xl)/2.0;
   ym = yl + (yr-yl)/2.0;
/* printf("M(xm,ym)[(%18.16f,%18.16f) (%18.16f,%18.16f)]=(%18.16f,%18.16f)\n",
	xl, yl, xr, yr, xm, ym );	/**/

   coordTransform( xm, ym, &Rm, &tm );
   /* printf("Painting (%3d,%3d)\n", Rm, tm );	/**/
   P[Rm][tm] = ON;

   fillPolarCurve( xl, yl, xm, ym, Rl, tl, Rm, tm, P );
   fillPolarCurve( xm, ym, xr, yr, Rm, tm, Rr, tr, P );
 }
}


/*** transforming from Cartesian to polar by representing line in 
*** parameter form and generating curve by this parameter 
***/
static void intpolPolarCurve( x1, y1, x2, y2, P )
int x1, y1, x2, y2;
Pixel **P;
{
 int R1, R2, t1, t2;

 /* Paint the two extremes */
 coordTransform( (float)x1, (float)y1, &R1, &t1 );
 P[R1][t1] = ON;
 coordTransform( (float)x2, (float)y2, &R2, &t2 );
 P[R2][t2] = ON;

 /*printf("INTERPOLATING BETWEEN (%3d,%3d) and (%3d,%3d)...\n",x1,y1,x2,y2);/**/
 fillPolarCurve((float)x1, (float)y1, (float)x2, (float)y2, R1, t1, R2, t2, P );
}


/****  Transformation :
***   for every Cartesian point calculate the corresponding
***   polar coordinate
***   CONNNECT NEIGHBOURS !!!
***/
void UNL_Transform( P, TP )
Pixel **P, **TP;
{
 int r, t, x, y;
 int xMin, xMax, yMin, yMax;
 int p00, p01, p10, p11;

 calcCentroid( P );
 clearPattern( TP );

 /* Set the bounding box for the normalized pattern */
 xMin = ROUND(O.x - maxDist) - 2;
 if( xMin < 0 )
   xMin = 0;
 xMax = ROUND((O.x + maxDist)) + 2;
 if(xMax > (N-1))
   xMax= N-1;
 yMin = ROUND((O.y - maxDist)) - 2;
 if(yMin < 0)
   yMin= 0;
 yMax = ROUND((O.y + maxDist)) + 2;
 if(yMax > (N-1))
   yMax= N-1;

 /* scan the image with a 2x2 mask and check for neighbors */
 for( y = yMin; y < yMax-1; y++ )
   for( x = xMin; x < xMax-1; x++ )
   {
     p00 = (int)P[x][y];
     p01 = (int)P[x][y+1];
     p10 = (int)P[x+1][y];
     p11 = (int)P[x+1][y+1];
    
     if( p00 && p10 )
       intpolPolarCurve( x, y, x+1, y, TP ); 
     if( p00 && p01 )
       intpolPolarCurve( x, y, x, y+1, TP ); 
     if( p00 && p11 && !p01 && !p10 )
       intpolPolarCurve( x, y, x+1, y+1, TP ); 
     if( p01 && p10 && !p00 && !p11 )
       intpolPolarCurve( x, y+1, x+1, y, TP );
   }
}
